#include <iostream>
#include <thread>
#include <chrono>
#include <string>
#include <nanomsg/nn.h>
#include <nanomsg/reqrep.h>
#include "rpcInterface.pb.h"

using namespace std;
using namespace myrpc;

//不同cpu下int在内存里的顺序不一样，不能莽memcpy
inline int bytes_to_int(uint8_t *address) {
    return (address[0]<<24) + (address[1]<<16) + (address[2]<<8) + address[3];
}

// @class rpcWorker
// 来自于 由rpcInterface.proto生成的myrpc::rpcService 
// 这是服务器端处理rpc的地方
class rpcWorker : public myrpc::rpcService {
public:
    void add(::google::protobuf::RpcController* controller,
            const ::myrpc::rpcRequest* request,
            ::myrpc::rpcResponse* response,
            ::google::protobuf::Closure* done) {
        cout<<" add function."<<endl;
        char result[128];
        sprintf(result, "%d+%d=%d", request->a(), request->b(), request->a()+request->b());
        response->set_result(result);
        this_thread::sleep_for( chrono::seconds(2) );
        cout<<" finish add function."<<endl;
        };
};

// @function showProtobufInfo
// 可以看 protobuf 文件里定义的server 的信息
// @input : The implement of my rpc servcie. (A service is a boundle of rpc functions).
void showProtobufInfo(google::protobuf::Service *serv) {
    cout<<"::Protobuf service Info::"<<endl;
    const google::protobuf::ServiceDescriptor *s_descriptor = serv->GetDescriptor();
    for (int i=0; i<s_descriptor->method_count(); i++) {
        const google::protobuf::MethodDescriptor *method = s_descriptor->method(i);
        cout<<"method "<<i<<" => "<<method->full_name()<<endl;
        cout<<"\t Input: "<<serv->GetRequestPrototype(method).GetTypeName()<<endl;
        cout<<"\tOutput: "<<serv->GetResponsePrototype(method).GetTypeName()<<endl;
    }
    cout<<endl;
}

// @function processMessage
// 收到了包，交给protobuf继承的rpc处理。
// 数据包格式：firstly sizeof(int) for function id, then other will be explained to myrpc::rpcRequest
// @param : 传入数据的地址，处理完把数据写回去
// @return : 处理完的buffer大小
int processMessage(google::protobuf::Service *serv, char *input) {
    const google::protobuf::ServiceDescriptor *s_desc = serv->GetDescriptor();
    int function_id = bytes_to_int((uint8_t*)input);
    cout<<"[Server] Function "<<function_id<<endl;
    if (function_id < 0 || function_id >= s_desc->method_count()) {
        cout<<"\t No function id found."<<endl;
        return 0;
    }

    // function_id 对应proto文件里 function的id
    const google::protobuf::MethodDescriptor *method = s_desc->method(function_id);
    google::protobuf::Message *input_argument = serv->GetRequestPrototype(method).New(); // 传入rpc函数的值
    google::protobuf::Message *result = serv->GetResponsePrototype(method).New(); // 执行完rpc的返回值
    
    // input数据包为 sizeof(int)表示函数id + 其余部分解析为input_argument
    input_argument->ParseFromString( std::string(input+sizeof(uint32_t)) );
    serv->CallMethod(method, nullptr, input_argument, result, nullptr);
    
    // 回传数据包格式： 直接把result序列化传回去。
    string output = result->SerializeAsString();
    memcpy(input, output.c_str(), output.length());
    return output.length();
}


// RPC函数运行端（server），流程如下
// 先初始化nanomsg socket，开request-response模式。
// 再初始化rpc处理class rpcWorker;
// 然后从网络中收到封装好的数据包，解包数据。(8字节functionID+其余传入参数)
// 扔给processMessage()函数调rpc，运行完得到var result。
// 然后序列化result，processMessage()返回到while里面，再扔给nn_req端。
int main() {
    //init for nanomsg socket
    int fd = nn_socket(AF_SP, NN_REP);
    assert(fd >= 0);
    int ret = nn_bind(fd, "ipc://rpc_address.proto");
    assert(ret >= 0);
    //init for protobuf service
    auto *myservice = new rpcWorker();
    showProtobufInfo(myservice);

    while (true) {
        cout<<"--Round Begin-- Waiting for rpc calling.."<<endl;
        char buffer[1024];
        nn_recv(fd, buffer, sizeof(buffer), 0);
        cout<<"** New Message **"<<endl;
        
        int buffsize = processMessage(myservice, buffer);
        cout<<"** Finish RPC running **"<<endl;

        nn_send(fd, buffer, buffsize, 0);
        cout<<"--Round End--"<<endl<<endl;
    }
    return 0;
}
